/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <Windows.H>
#include <StdLib.H>
#include <StdIO.H>

#include "CStack.H"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CStack::CStack(CStack *pStackToCopy)
{
	InitializeCriticalSection(&CS);

	this->Lock();
	pStackToCopy->Lock();

	this->Stack.AllocationIncrement = pStackToCopy->Stack.AllocationIncrement;
	this->Stack.Ceiling = pStackToCopy->Stack.Ceiling;
	this->Stack.Item = NULL;
	this->Stack.Used = pStackToCopy->Stack.Used;

	if(!(this->Stack.Item = (STACKITEM *) calloc(this->Stack.Ceiling, sizeof(STACKITEM))))
	{
		memset(&this->Stack, 0, sizeof(this->Stack));
		this->UnLock();
		return;
	}

	size_t uAllocate;
	STACKITEM *pItem;
	for(unsigned uItem = 0; uItem < pStackToCopy->Stack.Used; uItem++)
	{
		pItem = &pStackToCopy->Stack.Item[uItem];
		uAllocate = pItem->Size;

		if(pItem->Type == STACK_ITEM_TYPE_STRING)
		{
			uAllocate += 1; //MBS NULL terminator.
		}
		else if(pItem->Type == STACK_ITEM_TYPE_WSTRING)
		{
			uAllocate += 2; //WCS NULL terminator.
		}

		this->Stack.Item[uItem].Value = calloc(uAllocate, sizeof(byte));
		this->Stack.Item[uItem].Size = pItem->Size;
		this->Stack.Item[uItem].Type = pItem->Type;

		memcpy_s(this->Stack.Item[uItem].Value, uAllocate, pItem->Value, pItem->Size);

		if(pItem->Type == STACK_ITEM_TYPE_STRING)
		{
			((unsigned char *)this->Stack.Item[uItem].Value)[pItem->Size] = '\0';
		}
		else if(pItem->Type == STACK_ITEM_TYPE_WSTRING)
		{
			((WCHAR *)this->Stack.Item[uItem].Value)[pItem->Size] = '\0';
		}
	}

	pStackToCopy->UnLock();
	this->UnLock();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CStack::CStack()
{
	InitializeCriticalSection(&CS);

	this->Lock();
	memset(&this->Stack, 0, sizeof(this->Stack));
	this->Stack.AllocationIncrement = 10;
	this->UnLock();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CStack::~CStack()
{
	this->Clear();
	DeleteCriticalSection(&CS);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

size_t CStack::Allocated(void)
{
	return this->Stack.Ceiling;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

size_t CStack::StackSize(void)
{
	return this->Stack.Used;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CStack::Lock(void)
{
	EnterCriticalSection(&CS);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CStack::UnLock(void)
{
	LeaveCriticalSection(&CS);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CStack::Clear(void)
{
	this->Lock();
	if(this->Stack.Item)
	{
		for(unsigned int iIndex = 0; iIndex < this->Stack.Used; iIndex++)
		{
			free(this->Stack.Item[iIndex].Value);
		}
		free(this->Stack.Item);
	}

	memset(&this->Stack, 0, sizeof(this->Stack));

	this->UnLock();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(unsigned __int64 u64Value)
{
	return this->Push((void *) &u64Value, sizeof(u64Value), STACK_ITEM_TYPE_UINT64);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(__int64 i64Value)
{
	return this->Push((void *) &i64Value, sizeof(i64Value), STACK_ITEM_TYPE_INT64);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(unsigned int uValue)
{
	return this->Push((void *) &uValue, sizeof(uValue), STACK_ITEM_TYPE_UINT);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(int iValue)
{
	return this->Push((void *) &iValue, sizeof(iValue), STACK_ITEM_TYPE_INT);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(float fValue)
{
	return this->Push((void *) &fValue, sizeof(fValue), STACK_ITEM_TYPE_FLOAT);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(double dValue)
{
	return this->Push((void *) &dValue, sizeof(dValue), STACK_ITEM_TYPE_DOUBLE);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(const char *sValue)
{
	return this->Push((void *)sValue, strlen(sValue), STACK_ITEM_TYPE_STRING);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(const char *sValue, size_t uLength)
{
	return this->Push((void *)sValue, uLength, STACK_ITEM_TYPE_STRING);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(const WCHAR *wcValue)
{
	return this->Push((void *)wcValue, wcslen(wcValue), STACK_ITEM_TYPE_WSTRING);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(const WCHAR *wcValue, size_t uLength)
{
	return this->Push((void *)wcValue, uLength, STACK_ITEM_TYPE_WSTRING);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


bool CStack::Push(const void *pValue, size_t uLength)
{
	return this->Push((void *)pValue, uLength, STACK_ITEM_TYPE_BINARY);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Push(const void *pValue, size_t uLength, unsigned short uType)
{
	this->Lock();
	if(this->Stack.Used == this->Stack.Ceiling)
	{
		this->Stack.AllocationIncrement += (this->Stack.AllocationIncrement / 2);
		this->Stack.Ceiling += this->Stack.AllocationIncrement;
		if(!(this->Stack.Item = (STACKITEM *) realloc(this->Stack.Item, this->Stack.Ceiling * sizeof(STACKITEM))))
		{
			this->UnLock();
			return false;
		}
	}

	size_t uAllocate = uLength;

	if(uType == STACK_ITEM_TYPE_STRING)
	{
		uAllocate += 1; //MBS NULL terminator.
	}
	else if(uType == STACK_ITEM_TYPE_WSTRING)
	{
		uAllocate += 2; //WCS NULL terminator.
	}

	//Move all values down one position in the stack.
	for(register size_t uIndex = this->Stack.Used; uIndex > 0; uIndex--)
	{
		this->Stack.Item[uIndex] = this->Stack.Item[uIndex - 1];
	}

	this->Stack.Item[0].Value = calloc(uAllocate, sizeof(byte));
	this->Stack.Item[0].Size = uLength;
	this->Stack.Item[0].Type = uType;

	memcpy_s(this->Stack.Item[0].Value, uAllocate, pValue, uLength);

	if(uType == STACK_ITEM_TYPE_STRING)
	{
		((unsigned char *)this->Stack.Item[0].Value)[uLength] = '\0';
	}
	else if(uType == STACK_ITEM_TYPE_WSTRING)
	{
		((WCHAR *)this->Stack.Item[0].Value)[uLength] = '\0';
	}

	this->Stack.Used++;

	this->UnLock();

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(unsigned __int64 *u64Value)
{
	return this->Pop((void *)u64Value);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(__int64 *i64Value)
{
	return this->Pop((void *)i64Value);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(unsigned int *uValue)
{
	return this->Pop((void *)uValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(int *iValue)
{
	return this->Pop((void *)iValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(float *fValue)
{
	return this->Pop((void *)fValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(double *dValue)
{
	return this->Pop((void *)dValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(char *sValue, size_t *uLength)
{
	return this->Pop((void *)sValue, uLength);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(WCHAR *wcValue, size_t *uLength)
{
	return this->Pop((void *)wcValue, uLength);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(void *pValue, size_t *uLength)
{
	unsigned short uType;
	return this->Pop(pValue, uLength, &uType);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(void *pValue)
{
	size_t uLength;
	unsigned short uType;
	return this->Pop(pValue, &uLength, &uType);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Pop(void *pValue, size_t *uLength, unsigned short *uType)
{
	this->Lock();
	if(this->Stack.Used)
	{
		STACKITEM *pItem = &this->Stack.Item[--this->Stack.Used];
		memcpy_s(pValue, pItem->Size, pItem->Value, pItem->Size);

		*uLength = pItem->Size;
		*uType = pItem->Type;

		free(pItem->Value);
		this->UnLock();
		return true;
	}

	this->UnLock();
	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKITEM *CStack::Peek(void)
{
	return &this->Stack.Item[this->Stack.Used - 1];
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(unsigned __int64 *u64Value)
{
	return this->Peek((void *)u64Value);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(__int64 *i64Value)
{
	return this->Peek((void *)i64Value);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(unsigned int *uValue)
{
	return this->Peek((void *)uValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(int *iValue)
{
	return this->Peek((void *)iValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(float *fValue)
{
	return this->Peek((void *)fValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(double *dValue)
{
	return this->Peek((void *)dValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(char *sValue, size_t *uLength)
{
	return this->Peek((void *)sValue, uLength);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(WCHAR *wcValue, size_t *uLength)
{
	return this->Peek((void *)wcValue, uLength);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(void *pValue, size_t *uLength)
{
	unsigned short uType;
	return this->Peek(pValue, uLength, &uType);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(void *pValue)
{
	size_t uLength;
	unsigned short uType;
	return this->Peek(pValue, &uLength, &uType);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(void *pValue, size_t *uLength, unsigned short *uType)
{
	this->Lock();
	if(this->Stack.Used)
	{
		STACKITEM *pItem = &this->Stack.Item[this->Stack.Used - 1];
		memcpy_s(pValue, pItem->Size, pItem->Value, pItem->Size);

		*uLength = pItem->Size;
		*uType = pItem->Type;

		this->UnLock();
		return true;
	}

	this->UnLock();
	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKITEM *CStack::Peek(size_t uIndex)
{
	return &this->Stack.Item[uIndex];
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, unsigned __int64 *u64Value)
{
	return this->Peek(uIndex, (void *)u64Value);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, __int64 *i64Value)
{
	return this->Peek(uIndex, (void *)i64Value);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, unsigned int *uValue)
{
	return this->Peek(uIndex, (void *)uValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, int *iValue)
{
	return this->Peek(uIndex, (void *)iValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, float *fValue)
{
	return this->Peek(uIndex, (void *)fValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, double *dValue)
{
	return this->Peek(uIndex, (void *)dValue);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, char *sValue, size_t *uLength)
{
	return this->Peek(uIndex, (void *)sValue, uLength);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, WCHAR *wcValue, size_t *uLength)
{
	return this->Peek(uIndex, (void *)wcValue, uLength);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, void *pValue, size_t *uLength)
{
	unsigned short uType;
	return this->Peek(uIndex, pValue, uLength, &uType);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, void *pValue)
{
	size_t uLength;
	unsigned short uType;
	return this->Peek(uIndex, pValue, &uLength, &uType);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CStack::Peek(size_t uIndex, void *pValue, size_t *uLength, unsigned short *uType)
{
	this->Lock();
	if(this->Stack.Used)
	{
		STACKITEM *pItem = &this->Stack.Item[uIndex];
		memcpy_s(pValue, pItem->Size, pItem->Value, pItem->Size);

		*uLength = pItem->Size;
		*uType = pItem->Type;

		this->UnLock();
		return true;
	}

	this->UnLock();
	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

